##³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³ Importing libraries ³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³##

import q_learning_multispheres_cfg as cfg
import os
import argparse
import pandas as pd
import numpy as np
import random
from matplotlib import pyplot as plt
import itertools


##³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³ Defining Functions ³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³##

# This function generates simulations for a given action and state, reads the csv file
# and returns the reward. Note that the swimmer is located at the center (0,0) for each
# simulation, so the reward is the center of mass 0 of the center sphere.



def generate_simulation_Nsphere_swimmer(N, state, action):
    parser = argparse.ArgumentParser(description='Compute the .cfg, .json, .geo and preconditioner files for a chosen N-sphere swimmer.')
    parser.add_argument("--R",help="Radius of the spheres",type=float,default=1.0)
    parser.add_argument("--L",help="Distance between the centers of a pair of adjacent spheres",type=float,default=10.0)
    parser.add_argument("--Vel",help="Relative velocity between a pair of adjacent spheres",type=float,default=4.0)
    parser.add_argument("--Dim",help="The dimension",type=int,default=2)
    args = parser.parse_args()
    os.makedirs('../sw/cases/q-learning_multisphere', exist_ok=True)    
    cfg.write_cfg(N, '../sw/cases/q-learning_multisphere', args.Dim)
    cfg.write_json(N, action, '../sw/cases/q-learning_multisphere', args.Dim, args.Vel)
    cfg.write_geo(N, state,'../sw/cases/q-learning_multisphere', args.Dim, args.L, args.R)
    cfg.write_preconditioner(N, '../sw/cases/q-learning_multisphere', args.Dim)
    os.system('mpirun -np 8 feelpp_toolbox_fluid --config-file /ssd/esssousy/feelpp/toolboxes/fluid/moving_body/sw/cases/q-learning_multisphere/'+str(N)+'_sphere.cfg')
    data = pd.read_csv("/home/u2/essousy/feel/toolboxes/fluid/moving_body/"+str(N)+"_sphere/np_8/fluid.measures.csv")
    data.columns = data.columns.str.strip()
    sphere0_position = list(data['Quantities_body_Sphere0.mass_center_0'])
    reward = sphere0_position[len(sphere0_position)-1] - sphere0_position[0]
    return reward



#generate_simulation(N, state)
# This function takes for argument a state and an action, and returns the new state
# of the swimmer after taking the action in that state.

def new_state(state, action):
    New_State = [state[j] for j in range(0, len(state))]
    for i in range(0,len(state)):
        if state[i] == True and action == "retract_arm%s"%i :
            New_State[i] = not state[i]
        if state[i] == False and action == "extend_arm%s"%i :
            New_State[i] = not state[i]
    return New_State





# This function takes for argument a state and return the possible action the swimmer
# can perform. 

def possible_actions(state):
    actions = []
    for i in range(0, len(state)):
        if state[i]:
            actions.append("retract_arm%s"%i)
        else :
            actions.append("extend_arm%s"%i)
    return actions


# This function takes as arguments a dictionary and a value from it and returns the
# key of this value. Actually we are working with dictionaries for the actions and
# the states

def get_key(val, dict):
    for key, value in dict.items():
        if val == value:
            return key



# This function will be needed to access the keys of the possible actions. It is
# defined here to make the part of the Q-learning algorithm short and obvious.

def get_index_of_possible_actions(state):
    indexes = []
    for action in possible_actions(state):
        indexes.append(get_key(action, actions_space))
    return indexes




# This function will be used to determine the argmax of the Q-table. The Q-table 
# has 4 lines and 4 columns but the argmax should be over the possible actions and
# not all the actions. Hence, the numpy.argmax can't be used here since we take the
# argmax of specific values (possible actions). At each state, there are only two
# possible actions so we put their indexes here with the Q_table (to access the
# Q-table values) and this function returns the index of the action maximizing the
# Q-table.

def get_max_Q_table_index(possible_action_indexes, state_index, Q_table):
    Q_table_possible_actions = {possible_action_indexes[i] : Q_table[state_index, possible_action_indexes[i]] for i in range(0, len(possible_action_indexes))}
    Q_table_possible_actions_list = [Q_table[state_index, possible_action_indexes[i]] for i in range(0, len(possible_action_indexes))]
    Q_table_max = max(Q_table_possible_actions_list)
    index_max = get_key(Q_table_max, Q_table_possible_actions)
    return index_max



#########################################################################################

N = 5

##³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³ Define actions ³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³##

actions_space = {}
for i in range(0,N-1):
    actions_space[i] = "retract_arm%s"%i
for i in range(0,N-1):
    actions_space[i+N-1] = "extend_arm%s"%i
actions_list = [actions_space[i] for i in range(0, 2*N-2)]


##³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³ Define actions ³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³##

arm_states = [True, False]
states_list = [list(p) for p in itertools.product(arm_states, repeat=N-1)]
states_space = {}
for i in range(0, len(states_list)):
    states_space[i] = states_list[i]



# Names of states
# Here, there are 4 states, each state is a list of two components, each component
# is a boolien (True or Fals) describing the state of an arm. the first component is
# for the left arm and the second component is for the right arm, with the CONVENTION
# "False" means the arm is short (already retrated) and "True" means the arm is long.



##³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³ Q-learning algorithm ³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³³##


n = len(states_space)
m = len(actions_space)      # The Q-table dimensions (4 actions and 4 states)
alpha = 1                   # The learning rate for Q-learning algorithm
gamma = 0.99                # Discount factor for Q-learning algorithm
eps   = 0.1                 # Epsilon-greedy scheme constant
N_learn = 8000              # The number of learning steps


state = random.choice(states_list)  # The initial state of the swimmer will be state1=[True, True]
cumul_reward = 0                    # Cumulative reward initialisation
cumul_reward_list = []              # List that will contain cumulative reward of each learning step to be plotted at the end
actions_list = []                   # List that will contain the actions chosen over the learning process
states_list = []                    # List that will contain the states chosen over the learning process, starting from "state1" we have chosen
Q_table = np.zeros((n, m))          # Q-table initialisation by zeros
stored_results = np.zeros((n, m))   # Matrix of dimensions 4*4 to stores results of simulations to avoid repeating them!
# As this matrix is initialised with zeros, the reward of a simulation (state, action) will be put in the state_indes
# line and the action index column.

############################## 3-sphere swimmer stored rewards ##############################

path = "/home/u2/essousy/feel/toolboxes/fluid/moving_body/q-learning_multispheres_simulation_results_"+str(N)+"_SS.csv"
if os.path.isfile(path):
    reader = pd.read_csv(path)
    stored_results = np.array(reader).astype("float")[:, 1:]

print(stored_results)



for i in range(0, N_learn):
    #if i < 20:
    #    eps = 0.5
    #else :
    #    eps = 0.05
    state_index = get_key(state, states_space)     # Get the index of the state to be used to access Q-table
    # Epsilon-greedy scheme
    if random.uniform(0,1) < eps : 
        action = random.choice(possible_actions(state)) # Here we choose a random action from the possible actions!
        action_index = get_key(action, actions_space)   # Get the index of the action to be used to access Q-table
    else :                              # Here, the action that maximizes the Q-table will be chosen
        possible_action_indexes = get_index_of_possible_actions(state)   # Get indexes of possible actions so we can access their Q-table values to choose the maximum
        action_index = get_max_Q_table_index(possible_action_indexes, state_index, Q_table) # the function returns the index of the action maximizing the Q-table 
        action = actions_space[action_index]   # here, we return the action's name (it will be used to generate the simulation as the function (generating simulations) takes for argument the name and not the index)
    print("For i =", i, ":")
    print("state = ", state)
    print("action = ", action)
    if stored_results[state_index, action_index] == 0:  # if it is <0, it means we don't have results for this (state, action) situation. Hence, we need to perform the simulation
        reward = generate_simulation_Nsphere_swimmer(N, state, action)     # Generate the simulation and compute the reward
        New_state = new_state(state, action)            # The new state the swimmer moves to, after taking the action in the old state
        stored_results[state_index, action_index] = reward # Since we performed a new simulation, we store the reward so we don't need to perform it again
    else :                                              # Here (if =0) means the simulation for this case (state, action) has been performed and the reward has been stored in the matrix!
        reward = stored_results[state_index, action_index] # The reward has already been stored in the matrix
        New_state = new_state(state, action)               # The new state 
    #print("New_state = ", New_state)
    print("reward = ", reward)
    New_state_index = get_key(New_state, states_space)     # get the index of the state to be used to access Q-table
    New_indexes = get_index_of_possible_actions(New_state)
    Next_possible_actions_indexes = [Q_table[New_state_index, j] for j in New_indexes]
    max_value = max(Next_possible_actions_indexes)
    # The Q-learning formula
    Q_table[state_index, action_index] = Q_table[state_index, action_index] + alpha*(reward + gamma*max_value - Q_table[state_index, action_index])
    print(Q_table[New_state_index, :])
    #print("max_Q_table = ", max_value)
    cumul_reward += reward                  # Compute the cumulative reward
    print("cumul reward = ", cumul_reward)
    state = New_state                       # The state of the swimmer changes to the new one
    cumul_reward_list.append(cumul_reward)  # Add the cumulative reward to the list of cumulative rewards
    actions_list.append(action)             # Add the action to the list of actions
    states_list.append(New_state)           # Add the state to the list of states

if not os.path.isfile(path):
    pd.DataFrame(stored_results).to_csv(path)


#reader = pd.read_csv(path)
if not (stored_results == np.array(reader).astype("float")[:, 1:]).all() :
    pd.DataFrame(stored_results).to_csv(path)


print("Q-table = ", Q_table)
print("stored_results = ", stored_results)




